'use strict';

exports.__esModule = true;
exports.makeTree = makeTree;
exports.filterRedundantErrors = filterRedundantErrors;
exports.createErrorInstances = createErrorInstances;

var _utils = require('./utils');

var _validationErrors = require('./validation-errors');

var _debug = require('./debug');

const JSON_POINTERS_REGEX = /\/\w+(\/\d+)?/g;

// Make a tree of errors from ajv errors array
function makeTree(ajvErrors = []) {
  const root = { children: {} };
  ajvErrors.forEach(ajvError => {
    const { dataPath, keyword } = ajvError;

    // `dataPath === ''` is root
    const paths = dataPath === '' ? [''] : dataPath.match(JSON_POINTERS_REGEX);
    paths && paths.reduce((obj, path, i) => {
      obj.children[path] = obj.children[path] || { children: {}, errors: [] };
      if (i === paths.length - 1) {
        obj.children[path].errors.push(ajvError);
      }
      return obj.children[path];
    }, root);
  });
  return root;
}

function filterRedundantErrors(root, parent, key) {
  /**
   * If there is an `anyOf` error that means we have more meaningful errors
   * inside children. So we will just remove all errors from this level.
   */
  if ((0, _utils.getErrors)(root).some(_utils.isAnyOfError)) {
    root.errors = undefined;
  }

  /**
   * If all errors are `enum` and siblings have any error then we can safely
   * ignore the node.
   *
   * **CAUTION**
   * Need explicit `root.errors` check because `[].every(fn) === true`
   * https://en.wikipedia.org/wiki/Vacuous_truth#Vacuous_truths_in_mathematics
   */
  if (root.errors && (0, _utils.getErrors)(root).every(_utils.isEnumError)) {
    if ((0, _utils.getSiblings)(parent)(root)
    // Remove any reference which becomes `undefined` later
    .filter(_utils.notUndefined).some(_utils.getErrors)) {
      parent.children[key] = undefined;
    }
  }

  Object.entries(root.children).forEach(([key, child]) => filterRedundantErrors(child, root, key));
}

function createErrorInstances(root, indent) {
  const errors = (0, _utils.getErrors)(root);
  if (errors.length && errors.every(_utils.isEnumError)) {
    const allowedValues = (0, _utils.concatAll)([])(errors.map(e => e.params.allowedValues));
    const error = errors[0];
    return [new _validationErrors.EnumValidationError(Object.assign({}, error, {
      params: { allowedValues }
    }), indent)];
  } else {
    return (0, _utils.concatAll)(errors.reduce((ret, error) => {
      switch (error.keyword) {
        case 'additionalProperties':
          return ret.concat(new _validationErrors.AdditionalPropValidationError(error, indent));
        case 'required':
          return ret.concat(new _validationErrors.RequiredValidationError(error, indent));
        default:
          return ret.concat(new _validationErrors.DefaultValidationError(error, indent));
      }
    }, []))((0, _utils.getChildren)(root).map(child => {
      return createErrorInstances(child, indent);
    }));
  }
}

exports.default = (ajvErrors, indent) => {
  const tree = makeTree(ajvErrors || []);
  // debug(JSON.stringify(tree));
  filterRedundantErrors(tree);
  const errors = createErrorInstances(tree, indent);
  if (!errors) {
    (0, _debug.debug)(JSON.stringify(tree));
  }
  return errors;
};